# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT

"""Tests for argument parsing"""

import pytest

from tartex.__about__ import __version__
from tartex.tartex import TarTeX, make_tar


class TestArgs:
    """Class to test different combinations of cmdline arguments"""

    def test_no_args(self):
        """Test system exit with missing positional arg"""
        with pytest.raises(SystemExit) as exc:
            make_tar()

        # Weird: `hatch test -v` gets `exc.value.code = 1`, but `hatch test`
        # (without `-v`) gets `2`.  Actual value must be 1 --- there is no
        # `sys.exit(2)` in code, but help `hatch test` out.
        assert exc.value.code >= 1

    def test_only_file(self, sample_texfile):
        """Test success with one arg: file name"""
        assert sample_texfile.main_file.stem == "some_file"
        assert (
            sample_texfile.tar_file_w_ext.name
            == f"some_file.tar.{sample_texfile.tar_ext}"
        )

    def test_version(self, capsys, join_linebreaks):
        """Test version string against version from __about.py__"""
        # argparse will call SystemExit(0) for -h and -v, and print to stdout
        with pytest.raises(SystemExit) as exc:
            TarTeX(["--version"])

        output = join_linebreaks(capsys.readouterr().out)
        assert f"{__version__}" in output
        assert exc.value.code == 0

    @pytest.mark.parametrize(
        ("tar_opt1", "tar_opt2"), [("-J", "-z"), ("-j", "-J"), ("-z", "-J")]
    )
    def test_taropts_conflict(self, capsys, tar_opt1, tar_opt2):
        """Test exit status when two conflicting tar options are passed"""
        with pytest.raises(SystemExit) as exc:
            TarTeX([tar_opt1, tar_opt2, "some_file.tex"])

        assert exc.value.code == 2
        output = capsys.readouterr().err
        assert "not allowed with" in output



# module _tar
"""
Helper class TarFiles
"""

import logging as log
import math
import os
import sys
from io import BytesIO
from pathlib import Path
from rich import print as richprint
from .utils.tar_utils import TAR_EXT
from .utils.msg_utils import summary_msg
import tarfile as tar


class Tarballer:
    """Class that handles tarballing a list of objects (file Paths, BytesIO, etc.)"""

    def __init__(
        self, curr_dir: Path, main_input_file: Path, target: Path = Path(".")
    ):
        """Init class for TarFiles"""

        # set of actual, accessible files to include
        self._files: set[Path] = set()

        # dict of objects to include whose contents are only available as BytesIO
        # key: name to use; val: object contents as BytesIO
        self._streams: dict[str, BytesIO] = {}

        # dict of strings to use for logging when adding `key` object to tarball
        # key: file or stream object name; val: logging string
        self._comments: dict[str, str] = {}

        self._main_file: Path = main_input_file
        self._mtime = os.path.getmtime(self._main_file)

        self._work_dir: Path = curr_dir
        target_path: Path = self._work_dir / target
        self._target: Path
        self._ext = target_path.suffix.lstrip(".")
        self._target = target_path.with_suffix(f".{self._ext}")

    @property
    def num_objects(self) -> int:
        """Returns the total number of files and bytes objects added to tarball
        :returns: int

        """
        return len(self._files) + len(self._streams)

    def recomp_mode(self, recomp: str):
        """Set re-compression mode for tarball

        :recomp: one of "bz2", "gz", or "xz" (str)
        :returns: None

        """
        self._ext = recomp if recomp in TAR_EXT else "gz"
        self._target = self._target.with_suffix(self._ext)

    def set_mtime(self, mtime: float):
        """Set mtime for bytesio objects, for which mtime is not available otherwise"""
        self._mtime = mtime

    def app_files(self, *args: Path, comm: str = ""):
        """Update set of files to add to tarball with args

        :*args: files to append to `self._files` as args
        :returns: None

        """
        self._files.update(args)
        for f in args:
            if comm:
                self._comments[f.as_posix()] = comm

    def drop_files(self, *args: Path):
        """Drop files and associated comments, if any"""
        for f in args:
            self._files.discard(f)
            _ = self._comments.pop(f.as_posix(), None)

    def files(self) -> set[Path]:
        """Return copy of self._files as a set of Path(s)
        :returns: set of Paths

        """
        return self._files.copy()

    def streams(self) -> set[str]:
        """Return copy of self._streams.keys() as a set of strings
        :returns: set of str

        """
        return set(self._streams.keys())

    def objects(self) -> set[Path]:
        """Return all files and streams to be added to tarball"""
        return self.files().union([Path(f) for f in self.streams()])

    def app_stream(self, name: str, content: BytesIO, comm: str = ""):
        """Append a single BytesIO content to list of streams

        :name: file name to use for stream (str)
        :content: stream content corresponding to `name` (BytesIO)
        :returns: None

        """
        self._streams[name] = content
        if comm:
            self._comments[name] = comm

    def do_tar(self):
        def _tar_add_file(
            tar_obj: tar.TarFile,
            file_name: Path,
        ):  # helper func to add file <file_name> to <tar_obj>
            tinfo = tar_obj.gettarinfo(file_name)
            tinfo.uid = tinfo.gid = 0
            tinfo.uname = tinfo.gname = ""
            tar_obj.addfile(tinfo, open(file_name, "rb"))
            log.info("Add file: %s", file_name)

        def _tar_add_bytesio(tar_obj: tar.TarFile, file_name: str, obj):
            # helper func to add `obj` as BytesIO with filename <file_name> to <tar_obj>
            tinfo = tar_obj.tarinfo(file_name)
            tinfo.size = len(obj)
            tinfo.mtime = self._mtime
            tinfo.uid = tinfo.gid = 0
            tinfo.uname = tinfo.gname = ""
            tar_obj.addfile(tinfo, BytesIO(obj))
            log.info("Add contents as BytesIO: %s", file_name)

        # At this point, output tar name conflicts, if any, has been
        # resolved one way or another, and we should simply over-write
        # tarball if it exists. So, it is safe to use 'w:'
        try:
            lof = self._files.copy()
            with tar.open(self._target, mode=f"w:{self._ext}") as tar_obj:
                for dep in lof:
                    try:
                        _tar_add_file(tar_obj, dep)
                    except FileNotFoundError:
                        log.warning(
                            "Skip missing INPUT file '%s'; try"
                            " forcing a LaTeX recompile ('-F').",
                            dep,
                        )
                        self._files.discard(dep)
                        continue
                for key, val in self._streams.items():
                    _tar_add_bytesio(tar_obj, key, val)
        except PermissionError as e:
            log.critical(e)
            sys.exit(1)
        except Exception as e:
            raise e

    def print_list(self, _summ: bool = False):
        """helper function to print list of files in a pretty format"""
        total = len(self._files) + len(self._streams)
        idx_width = int(math.log10(total)) + 1  # num of chars for serial nums
        files_cnt = 0
        for f in sorted(self._files):
            if f.exists():
                richprint(f"{files_cnt + 1:{idx_width}}.", end=" ")
                print(str(f))
                files_cnt += 1
            else:
                log.warning("Missing file skipped: %s", f)
                total -= 1
        for r in sorted(self._streams.keys()):
            richprint(f"{'*':>{idx_width + 1}}", end=" ")
            print(f"{r}")
        if _summ:
            summary_msg(total)



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT

"""Common fixtures"""

import os
from pathlib import Path
import shutil
import time
import pytest

from tartex.tartex import TarTeX
import tartex.utils.xdgdir_utils as _tartex_xdg_utils

@pytest.fixture(autouse=True)
def mock_home(monkeypatch, tmp_path):
    monkeypatch.setattr(Path, "home", lambda: tmp_path)


@pytest.fixture(autouse=True)
def mock_cache_dir(monkeypatch, tmp_path):
    monkeypatch.setattr(
        _tartex_xdg_utils,
        "XDG_CACHE_HOME",
        tmp_path / "cache",
    )


@pytest.fixture
def monkeypatch_mtime(monkeypatch):
    return lambda _: monkeypatch.setattr(
        "os.path.getmtime", lambda _: time.time()
    )


@pytest.fixture
def monkeypatch_set_main_file(monkeypatch):
    return lambda foo: monkeypatch.setattr(
        "tartex.tartex._set_main_file", lambda _: Path(foo)
    )


@pytest.fixture
def sample_texfile(monkeypatch_set_main_file, monkeypatch):
    """Pytest fixture: TarTeX with just a tex file for parameter"""

    # "some_file.tex" does not actually exist, but we use the dummy tartex
    # object to test class member variables anyway.
    #
    # For this dummy object to not cause early exit of TarTeX, we monkeypatch
    # `_set_main_file` to just return a path to the non-existent
    # "some_file.tex" instead of returning None. The latter would cause the
    # TarTeX object to hit sys.exit during __init__(), raising test errors.
    def mock_time(_):
        return time.time()

    monkeypatch.setattr("os.path.getmtime", mock_time)
    monkeypatch_set_main_file("some_file.tex")
    return TarTeX(["some_file.tex"])


# Data copying to tmpdir to allow using with pytest
# Note that relative paths to the data do not work as it is never certain which
# location pytest is run from
# Based on https://stackoverflow.com/a/29631801
@pytest.fixture
def datadir(tmpdir, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        shutil.copytree(test_dir, tmpdir, dirs_exist_ok=True)

    return tmpdir


@pytest.fixture
def join_linebreaks():
    """Joins lines with line breaks into one contiguous string

    :returns: Lambda that joins line at line breaks (str)

    """
    return lambda lbreaks: "".join(lbreaks.split("\n"))



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#
"""Tests for tarball generation from basic latex files"""

import json
import logging
import os
import tarfile as tar
from pathlib import Path
from tartex.utils.hash_utils import HASH_METHOD

import pytest

from tartex.tartex import TarTeX


@pytest.fixture
def default_target(datadir):
    return Path(datadir) / "test"


@pytest.fixture
def default_tartex_obj(datadir, default_target):
    return lambda fname: TarTeX(
        [
            (Path(datadir) / fname).as_posix(),
            "-v",
            "-s",
            "-o",
            default_target.as_posix(),
        ]
    )


class TestBasicLaTeX:
    """Tests checking tar file generation from a basic latex file"""

    def test_list_only(self, datadir, default_target, capsys):
        """Test `--list` and `--dry-run` options which _do not_ produce tarballs"""
        output = default_target.with_suffix(".tar.gz")
        for opt in ["-l", "--list", "--dry-run"]:
            t = TarTeX([(Path(datadir) / "basic_latex.tex").as_posix(), opt])
            t.tar_files()
            assert t.args.list
            assert not output.exists()
            assert "basic_latex.tex" in capsys.readouterr().out.strip()

    def test_gen_tar(self, default_target, default_tartex_obj):
        """Should include a single file in tarball"""
        output = default_target.with_suffix(".tar.gz")
        t = default_tartex_obj("basic_latex.tex")
        t.tar_files()
        assert output.exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_gen_tar_no_suffix(self, default_target, default_tartex_obj):
        """Should include a single file in tarball even though main file has no
        .tex or .fls suffix
        """
        output = default_target.with_suffix(".tar.gz")
        t = default_tartex_obj("basic_latex")
        t.tar_files()
        assert output.exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_diff_target_dir(self, tmpdir, datadir):
        # Make a new dir inside tmpdir
        destdir = tmpdir / "dest"
        os.mkdir(destdir)
        t = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "-v",
                "-s",
                "-o",
                str(destdir / "output.tar.gz"),
            ]
        )
        t.tar_files()
        assert t.tar_file_w_ext.exists()

    def test_incl_pdf(self, datadir):
        """
        Test option `--with-pdf` for inclusion of pdf in tarball
        """
        t = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "--with-pdf",
                "-o",
                f"{datadir}/basic_latex_with_pdf.tar.gz",
            ]
        )
        t.tar_files()
        with tar.open(t.tar_file_w_ext) as f:
            assert len(f.getnames()) == 2
            assert t.main_file.with_suffix(".pdf").name in f.getnames()

    def test_verbose_debug(self, datadir):
        """
        Test verbose value set by -vv option
        """
        t = TarTeX([(Path(datadir) / "basic_latex.tex").as_posix(), "-vv"])
        assert t.args.verbose == 2

class TestCache:
    """Tests involving cache generation, retrieval, and update"""

    def test_cache_gen(self, datadir):
        """
        Test cache file is generated when tartex is first run
        """
        t = TarTeX(
            [
                str(datadir / "basic_latex.tex"),
                "-v",
                "-o",
                str(datadir / "cache_gen.tar.gz"),
            ]
        )
        t.tar_files()
        assert t.tar_file_w_ext.exists()
        assert t.filehash_cache.exists()
        with open(t.filehash_cache, "r") as _c:
            _j = json.load(_c)
            assert "basic_latex.tex" in _j["input_files"]
            assert not _j["streams"]  # _j["streams"] is empty list

    def test_cache_retrieve(self, datadir):
        """
        Test cache file is untouched when tartex is rerun without changes to
        input files
        """
        t = TarTeX(
            [
                str(datadir / "basic_latex.tex"),
                "-v",
                "--overwrite",
                "-o",
                str(datadir / "cache_ret.tar.gz"),
            ]
        )
        t.tar_files()
        assert t.filehash_cache.exists()
        mtime = int(os.path.getmtime(t.filehash_cache))
        t.tar_files()  # run mustn't touch cache, as no input files changed
        assert int(os.path.getmtime(t.filehash_cache)) == mtime

    def test_cache_update(self, datadir):
        """
        Test cache file is updated when tartex is run after input files change
        """
        t = TarTeX(
            [
                str(datadir / "basic_latex.tex"),
                "-v",
                "--overwrite",
                "-o",
                str(datadir / "cache_up.tar.gz"),
            ]
        )

        # First run
        t.tar_files()
        assert t.filehash_cache.exists()
        mtime = os.path.getmtime(t.filehash_cache)
        with open(t.filehash_cache, "r") as _c:
            _j = json.load(_c)

        # Modify file
        with open(t.main_file, "a") as _f:
            _f.write("\n")

        # Second run
        t.tar_files()

        assert os.path.getmtime(t.filehash_cache) != mtime
        assert (
            _j["input_files"][t.main_file.name]
            != HASH_METHOD(t.main_file.read_bytes()).hexdigest()
        )
        with open(t.filehash_cache, "r") as _c:
            _j = json.load(_c)
        assert (
            _j["input_files"][t.main_file.name]
            == HASH_METHOD(t.main_file.read_bytes()).hexdigest()
        )

    def test_cache_force_recompile(self, datadir):
        """
        Test cache file is untouched when tartex is run with `-F`
        """
        t = TarTeX(
            [
                str(datadir / "basic_latex.tex"),
                "-v",
                "-F",
                "-o",
                str(datadir / "cache_recompile.tar.gz"),
            ]
        )
        t.tar_files()
        assert not t.filehash_cache.exists()

# These tests involve repeatedly compiling LaTeX files, thus can be slow
@pytest.mark.slow
class TestTarConflict:
    """Tests checking resolutions for tar file name conflicts"""

    def test_sol_default(
        self,
        default_target,
        default_tartex_obj,
        capsys,
        monkeypatch,
        join_linebreaks,
    ):
        """Test default (empty) user response"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "")

        # Trying to create tar file again will lead to conflict res dialog
        # Blank user input (from monkeypatch) will raise SystemExit
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        # Original output must still exist
        assert default_target.with_suffix(".tar.gz").exists() is True

        assert "Not overwriting existing tar file" in join_linebreaks(
            capsys.readouterr().err
        )
        assert exc.value.code == 1

    def test_sol_quit(
        self,
        default_target,
        default_tartex_obj,
        capsys,
        monkeypatch,
        join_linebreaks,
    ):
        """Test when user response is 'q'"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "q")

        # Trying to create tar file again will lead to conflic res dialog
        # Blank user input (from monkeypatch) will raise SystemExit
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        # Original output must still exist
        assert default_target.with_suffix(".tar.gz").exists() is True

        assert "Not overwriting existing tar file" in join_linebreaks(
            capsys.readouterr().err
        )
        assert exc.value.code == 1

    def test_sol_overwrite(self, default_tartex_obj, monkeypatch):
        """Test overwrite resolution"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        # Monkeypatch empty response for input
        monkeypatch.setattr("rich.prompt.Prompt.ask", lambda _: "o")
        t_con.tar_files()
        output = t_con.tar_file_w_ext
        assert output.exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_sol_newname_ok(
        self, default_target, default_tartex_obj, tmpdir, monkeypatch
    ):
        """Test entering new name that works"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        output = str(tmpdir / "new.tar.gz")
        # Monkeypatch responses for choosing a new file name
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        t_con.tar_files()

        # Original output must still exist
        assert default_target.with_suffix(".tar.gz").exists() is True

        assert Path(output).exists() is True
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_sol_newname_old(
        self, default_tartex_obj, tmpdir, capsys, monkeypatch, join_linebreaks
    ):
        """Test error when entering new name that is same as the old name"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        output = str(tmpdir / "test.tar.gz")

        # Monkeypatch responses for choosing file name same as original
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        with pytest.raises(SystemExit) as exc:
            t_con.tar_files()

        assert "New name entered is also the same" in join_linebreaks(
            capsys.readouterr().err
        )
        assert exc.value.code == 1
        # Original output must still exist
        assert Path(output).exists() is True

    def test_sol_newext(self, default_tartex_obj, tmpdir, monkeypatch):
        """Test new name with just the file ext changed"""
        t_con = default_tartex_obj("basic_latex.tex")
        t_con.tar_files()

        output = str(tmpdir / "test.tar.gz")
        assert Path(output).exists() is True

        output = output.replace(".gz", ".xz")
        # Monkeypatch responses for choosing file name same as original
        user_inputs = iter(["c", output])
        monkeypatch.setattr(
            "rich.prompt.Prompt.ask", lambda _: next(user_inputs)
        )
        t_con.tar_files()
        assert Path(output).exists() is True
        assert t_con.tar_ext == "xz"
        with tar.open(output) as rat:
            assert len(rat.getnames()) == 1

    def test_option_overwrite(
        self, caplog, datadir, default_target, default_tartex_obj
    ):
        """Should include a single file in tarball"""
        output = default_target.with_suffix(".tar.gz")
        t = default_tartex_obj("basic_latex.tex")
        t.tar_files()
        q = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "--overwrite",
                "-o",
                default_target.as_posix(),
            ]
        )
        q.tar_files()
        assert output.exists() is True
        caplog.set_level(logging.WARNING)
        out_mess = caplog.text
        assert "overwriting existing" in out_mess.lower()

class TestOnlyCheckBasic:
    def test_only_check_success(self, capsys, datadir):
        """Use `--only-check` to report successful inclusion of single file"""
        t = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "--only-check",
                "-o",
                (Path(datadir) / "basic_latex_check.tar.gz").as_posix(),
            ]
        )
        t.tar_files()
        assert not t.tar_file_w_ext.exists()
        assert "All files needed for compilation included" in capsys.readouterr().out

    def test_check_warn_pdf(self, capsys, datadir):
        """Warn when `--only-check` finds unnecessary files to be included"""
        t = TarTeX(
            [
                (Path(datadir) / "basic_latex.tex").as_posix(),
                "--only-check",
                "--with-pdf",
                "-o",
                (Path(datadir) / "basic_latex_warn.tar.xz").as_posix(),
            ]
        )
        t.tar_files()
        assert not t.tar_file_w_ext.exists()
        msg_out = capsys.readouterr().out
        assert "Files not essential to compilation" in msg_out
        assert "All files needed for compilation" in msg_out



# vim: set et ai ts=4 sw=4 tw=100:

import shutil
import subprocess
import logging as log
from pathlib import Path
from contextlib import contextmanager


# This is a context manager that modifies the user's git working tree in place, i.e., it checks out
# the specified revision (if valid) in a `--detached` mode, yields for work to be done on the
# working tree in this detached checkout mode, and cleans up by restoring the branch HEAD.
#
# While useful, this might lead to race effects if the user manages to change any files in the
# repository while the context manager is active. Instead, consider switching to using pure git refs
# from `ls-tree --long` and getting binary data from the refs using `git cat-file`. This approach
# will avoid modifying anything in the user dir at all, rather than the current checkout-and-restore
# approach.
@contextmanager
def git_checkout(git_bin: str, repo: str, rev: str):
    """
    A context manager for git checkout that checks out `rev` and restores previous working tree
    afterwards
    """

    head_abbrev_ref = _get_ref(git_bin, repo, "HEAD", abbrev=True)
    head_full_ref = _get_ref(git_bin, repo, "HEAD", abbrev=False)
    if head_abbrev_ref == "HEAD":
        # This means we are in a detached working tree to begin with
        # and will require the full hash to restore tree at the end
        head_short_ref = head_full_ref[:7]
    else:
        # HEAD points to branch tip
        head_short_ref = head_abbrev_ref

    try:
        rev_full_ref = _get_ref(git_bin, repo, rev)
        rev_short_ref = rev_full_ref[:7]
    except subprocess.CalledProcessError as err:  # typically for invalid rev
        log.critical(
            "Failed to checkout revision %s; invalid git revision?", rev
        )
        log.critical("Git: %s", err.stderr.strip())
        raise err

    if head_full_ref != rev_full_ref:
        log_tgt_rev = (
            rev if rev == rev_short_ref else f"{rev} (commit {rev_short_ref})"
        )
        try:
            # Note: git prints checkout msgs to stderr, not stdout
            rev_proc = subprocess.run(
                [git_bin, "-C", repo, "checkout", "--detach", rev],
                capture_output=True,
                encoding="utf-8",
                check=True,
            )

            log.info(
                "Checking out git revision %s in detached mode", log_tgt_rev
            )
            for line in rev_proc.stderr.splitlines():
                log.debug("Git: %s", line)

            yield rev_short_ref

        except subprocess.CalledProcessError as err:
            log.critical("Failed to checkout git revision %s", log_tgt_rev)
            for line in err.stderr.splitlines():
                log.critical("Git: %s", line)
            raise err

        restore_ref = (
            head_full_ref if head_abbrev_ref == "HEAD" else head_abbrev_ref
        )
        head_proc = subprocess.run(
            [git_bin, "-C", repo, "checkout", "-f", restore_ref],
            capture_output=True,
            encoding="utf-8",
            check=True,
        )
        log.info("Restoring git working tree to revision %s", head_short_ref)
        for line in head_proc.stderr.splitlines():
            log.debug("Git: %s", line)
    else:
        check_clean_git_tree = subprocess.run(
            [
                git_bin,
                "-C",
                repo,
                "status",
                "--porcelain",
                "--untracked-files=no",
            ],
            capture_output=True,
            encoding="utf-8",
            check=False,
        )
        log_msg_tree = (
            "(unclean)"
            if check_clean_git_tree.stdout
            else f"at {head_short_ref}"
        )
        log.info("Using current working tree %s", log_msg_tree)
        yield rev_short_ref


def _get_ref(git_bin, repo, rev, abbrev=False):
    """Returns the full ref for a given git revision `rev` in a git `repo`"""

    cmd = [git_bin, "-C", repo, "rev-parse"]
    if abbrev:
        cmd += ["--abbrev-ref"]
    return subprocess.run(
        [*cmd, rev],
        capture_output=True,
        encoding="utf-8",
        check=True,
    ).stdout.splitlines()[0]


class GitRev:
    """Class to set up and obtain file list from a git repo"""

    def __init__(self, repo: str, rev: str = "HEAD") -> None:
        """Initialise GitRev class

        :repo: dir containing a git repo (str)
        :rev: a valid git revision (str), default: "HEAD"

        """
        self.repo: str = str(repo)
        self.rev: str = rev
        self.git_bin = shutil.which("git")
        if not self.git_bin:
            raise RuntimeError("Unable to find git executable in PATH")

    def mtime(self) -> int:
        """Return the target commit's unix timestamp"""
        try:
            return int(
                self._git_cmd(
                    ["show", "--no-patch", r"--format=%ct", self.rev],
                )[0].strip()
            )
        except Exception as err:
            raise err

    def id(self) -> str:
        """Return either tag name, if commit corresponds to a valid tag, or commit short-id
        otherwise

        :returns: str

        """
        # The first line of the commit is something like the following:
        # ```
        # commit SHORT_ID
        # ```
        try:
            self.git_commit_id = self._git_cmd(
                [
                    "show",
                    "--abbrev-commit",
                    "--no-patch",
                    "--no-color",
                    self.rev,
                ]
            )[0]

        except subprocess.CalledProcessError as err:
            for line in err.stderr.splitlines():
                log.critical("Git: %s", line)
            raise err

        self.tag_id: str | None
        try:
            self.tag_id = self._git_cmd(
                [
                    "describe",
                    "--tags",
                    "--exact-match",
                    self.rev,
                ]
            )[0]
        except subprocess.CalledProcessError as err:
            log.info("Git: No tag corresponding to ref: %s", self.rev)
            for line in err.stderr.splitlines():
                log.debug("Git: %s", line)
            self.tag_id = None

        return self.tag_id or f"git.{self.git_commit_id.split()[1]}"

    def ls_tree_files(self) -> set[Path]:
        """Get list of files from ls-tree
        :returns: dict[Path]

        """
        try:
            _files = self._git_cmd(
                [
                    "ls-tree",
                    "-r",
                    "--name-only",
                    self.rev,
                ]
            )
            self.ls_tree_paths = set([Path(f) for f in _files])

        except (OSError, subprocess.CalledProcessError):
            self.ls_tree_paths = set()

        return self.ls_tree_paths

    def _git_cmd(self, cmd: list[str]) -> list[str]:
        """Run specified cmd and return captured output

        :cmd: git command to run (list[str])
        :returns: list of lines from output (list[str])
        :raises: OSError, subprocess.CalledProcessError

        """

        git_comm: list[str] = [
            self.git_bin if self.git_bin else "git",
            "-C",
            self.repo,
        ] + cmd
        try:
            out = subprocess.run(
                git_comm, capture_output=True, encoding="utf-8", check=True
            )
        except (OSError, subprocess.CalledProcessError) as err:
            raise err
        return out.stdout.splitlines()



# vim: set ai et ts=4 sw=4 tw=100:

import hashlib
import shutil
import pytest
from pathlib import Path
from subprocess import run
import tarfile as tar

from tartex.tartex import TarTeX


@pytest.fixture
def git_bin():
    """Returns the git executable available in PATH
    :returns: str pointing to git binary or None if `git` in not found in Path

    """
    return shutil.which("git")


@pytest.fixture
def git_repo_clean(datadir, git_bin):
    """Set up a clean git repository"""
    git_cmd_base = [git_bin, "-C", datadir]
    run([*git_cmd_base, "init", "-b", "main"])
    run([*git_cmd_base, "add", "."])
    run([*git_cmd_base, "commit", "-m", "Initial commit."])
    r = run(
        [*git_cmd_base, "rev-parse", "HEAD"],
        capture_output=True,
        encoding="utf-8",
    )
    return datadir, git_cmd_base, r.stdout.splitlines()[0].strip()


@pytest.fixture
def gitrev_tartex(datadir):
    return lambda rev: TarTeX(
        [
            (Path(datadir) / "git_rev").as_posix(),
            "-vv",
            "-s",
            "-o",
            Path(datadir).as_posix(),
            "-g",
            rev,
        ]
    )


class TestGitRev:
    """Test class for git-rev"""

    def test_default_tar_filename(self, git_repo_clean, datadir, gitrev_tartex):
        git_repo, git, git_ref = git_repo_clean
        assert git_repo == datadir
        assert git_ref
        git_short_ref = git_ref[:7]
        # r1 should tarball working tree as it appeared at first commit
        r1 = gitrev_tartex(git_ref)
        assert (
            r1.main_file.stem + f"-git.{git_short_ref}.tar.{r1.tar_ext}"
        ) == r1.tar_file_w_ext.name

    def test_tarfile_change_tag(self, git_repo_clean, datadir, gitrev_tartex):
        git_repo, git, git_ref = git_repo_clean
        git_short_ref = git_ref[:7]
        run(
            [
                "sed",
                "-i",
                "s/test document/Test Document/",
                Path(datadir) / "git_rev.tex",
            ]
        )
        run([*git, "commit", "-a", "-m", "Second commit"])

        # r2 should tarball working tree as it appears at current HEAD
        r2 = gitrev_tartex("")
        assert f".{git_short_ref}" not in Path(r2.tar_file_w_ext).suffixes

    def test_final_git_branch_head(
        self, git_repo_clean, datadir, gitrev_tartex
    ):
        """Test whether we restore git tree pristinely at branch tip"""
        git_repo, git, git_ref = git_repo_clean
        run(
            [
                "sed",
                "-i",
                "s/test document/Test Document/",
                Path(datadir) / "git_rev.tex",
            ]
        )
        run([*git, "commit", "-a", "-m", "Second commit"])
        # Now we are at "HEAD", pointing to "main" branch tip...
        gitrev_tartex(git_ref)
        git_status = run(
            [*git, "status"], capture_output=True, check=True, encoding="utf-8"
        )
        # ...first line should be "On branch main"
        assert "On branch" in git_status.stdout.splitlines()[0]

    def test_final_git_detach_head(
        self, git_repo_clean, datadir, gitrev_tartex
    ):
        """Test whether we restore git tree back to detached head if that is
        the initial status of the tree
        """
        git_repo, git, git_ref = git_repo_clean
        run(
            [
                "sed",
                "-i",
                "s/test document/Test Document/",
                Path(datadir) / "git_rev.tex",
            ]
        )
        run([*git, "commit", "-a", "-m", "Second commit"])
        git_ref_2 = run(
            [*git, "rev-parse", "HEAD"],
            capture_output=True,
            encoding="utf-8",
        ).stdout.strip()
        run([*git, "checkout", "--detach", git_ref])
        # Now we are at detached "HEAD", pointing to initial commit...
        gitrev_tartex(git_ref_2)
        git_status = run(
            [*git, "status"], capture_output=True, check=True, encoding="utf-8"
        )
        # ...first line should be "HEAD detached at"
        assert "HEAD detached at" in git_status.stdout.splitlines()[0]

    def test_tar_contents(self, git_repo_clean, datadir, gitrev_tartex, capsys):
        """validate `git_rev.tex` in tarball against original file"""

        git_repo, git, git_ref = git_repo_clean
        git_short_ref = git_ref[:7]
        with open(Path(datadir) / "git_rev.tex", "rb") as f:
            r1_tex_data = f.read()
            r1_texfile_sha = hashlib.sha1(r1_tex_data)

        run(
            [
                "sed",
                "-i",
                "s/test document/Test Document/",
                Path(datadir) / "git_rev.tex",
            ]
        )
        run([*git, "commit", "-a", "-m", "Second commit"])

        t = gitrev_tartex(git_short_ref)
        t.tar_files()

        with tar.open(t.tar_file_w_ext, "r") as tf:
            tex_file = tf.extractfile(t.main_file.name)
            if tex_file:
                tex_data = tex_file.read()
                assert tex_data.decode("utf-8") == r1_tex_data.decode("utf-8")
                tex_data_sha1 = hashlib.sha1(tex_data)
                assert tex_data_sha1.hexdigest() == r1_texfile_sha.hexdigest()



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#

from pathlib import Path
from typing import Union
import json

import re


# Auxilliary file extensions to ignore
# taken from latexmk manual:
# https://www.cantab.net/users/johncollins/latexmk/latexmk-480.txt
AUXFILES = [
    ".aux",
    ".bcf",
    ".fls",
    ".idx",
    ".lof",
    ".lot",
    ".out",
    ".toc",
    ".blg",
    ".ilg",
    ".log",
    ".xdv",
    ".fdb_latexmk",
]

# Supplementary files that are usually required as part of tarball
SUPP_REQ = ["bbl", "ind"]


def add_files(patterns: list[str], dir: Path) -> set[Path]:
    """
    Return list of additional user specified/globbed file paths,
    if they exist
    """
    files: set[Path] = set()
    for fpatt in patterns:
        files.update(set(dir.glob(f"**/{fpatt}")))

    return files


# TODO: account for possible multiple bib file usage
def bib_file(tex_fname: Path) -> list[Union[Path, None]]:
    """Return relative path to bib file"""
    bibre = re.compile(r"^\\bibliography\{.*\}")
    bib_name = None
    bstre = re.compile(r"^\\bibliographystyle\{.*\}")
    bst_name = None
    with open(tex_fname, encoding="utf-8") as f:
        for line in f:
            m_bib = bibre.search(line)
            m_bst = bstre.search(line)
            if m_bib:
                bib_name = m_bib.group()
                continue
            if m_bst:
                bst_name = m_bst.group()
                continue
            if bib_name and bst_name:
                break

    if bib_name:
        bib_name = re.sub(r"^\\bibliography\{", "", bib_name).rstrip("}")
        bib_name += ".bib" if bib_name.split(".")[-1] != ".bib" else ""
    if bst_name:
        bst_name = re.sub(r"^\\bibliographystyle\{", "", bst_name).rstrip("}")
        bst_name += ".bst" if bst_name.split(".")[-1] != ".bst" else ""

    return [Path(f) for f in [bib_name, bst_name] if Path(f).is_file()]

class SetEncoder(json.JSONEncoder):
    """A class to allow JSONEncoder to interpret a set as a list"""

    def default(self, o: set) -> list:
        """
        Convert o (a set) into a sorted list

        :o: set
        :return: list
        """
        return sorted(o)



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#

from pathlib import Path
from typing import Union
from rich import print as richprint


def summary_msg(
    nfiles, tarname: Union[Path, None] = None, wdir: Union[Path, None] = None
):
    """Return summary msg to print at end of run"""

    def _num_tag(n: int):
        return f"[bold]{n} file" + ("s" if n > 1 else "") + "[/]"

    if tarname:
        try:
            tarname_rel = tarname.relative_to(wdir if wdir else tarname.root)
        except ValueError:
            tarname_rel = tarname
        finally:
            richprint(
                f"[cyan]Summary: :package: [bold]{tarname_rel}[/] generated with"
                f" {_num_tag(nfiles)}.[/]"
            )
    else:
        richprint(
            f"[cyan]Summary: :clipboard: {_num_tag(nfiles)} to include.[/]"
        )



# vim:set et sw=4 ts=4:
# SPDX-FileCopyrightText: 2024-present Atri Bhattacharya <atrib@duck.com>
#
# SPDX-License-Identifier: MIT
#
"""tartex module"""

import fnmatch
import json
import logging as log
import os
import re
import sys
from contextlib import nullcontext
from pathlib import Path
from rich import (
    print as richprint,
    logging as richlog,
    highlighter,
)
from tempfile import TemporaryDirectory
from typing import Union

from tartex import _latex
from tartex._parse_args import parse_args
from tartex._git_rev import GitRev, git_checkout
import tartex.utils.msg_utils as _tartex_msg_utils
import tartex.utils.tex_utils as _tartex_tex_utils
import tartex.utils.tar_utils as _tartex_tar_utils
import tartex.utils.hash_utils as _tartex_hash_utils
import tartex.utils.xdgdir_utils as _tartex_xdg_utils
from tartex._tar import Tarballer

try:
    from contextlib import chdir
# contextlib.chdir is only available from Python 3.10 onwards
except ImportError:
    from contextlib import contextmanager

    @contextmanager  # type: ignore [no-redef]
    def chdir(p: Path):
        cwd = Path.cwd()
        try:
            os.chdir(p)
            yield p
        except Exception as err:
            raise err
        finally:
            os.chdir(cwd)

##################################
## HELPER CLASSES AND FUNCTIONS ##
##################################


class CheckFailError(Exception):
    """Exception raised when `--check` fails"""

    def __init__(self, msg=""):
        """init CheckFail exception"""
        super().__init__(msg)
        self._msg = msg

    def __str__(self):
        return f"Check failed: {self._msg}"


def _check_err_missing(
    _ref: Tarballer,
    _tgt: Tarballer,
    *,
    supp_files: set[Path] = set(),
    indicator: str = "*",
) -> bool:
    """
    Compares objects included in _ref and _tgt and return True if the former
    contains files missing from the latter. Logs accordingly.

    If non-empty, warns instead of declaring errors on missing `supp_files`.
    """
    non_exist_files = set([f for f in _ref.objects() if not f.exists()])
    non_exist_files.difference_update(supp_files)
    excluded = _ref.objects().difference(_tgt.objects())
    excluded.difference_update(non_exist_files.union(supp_files))

    if non_exist_files or excluded:
        if non_exist_files:
            log.info(
                "Files needed for compilation not found (deleted?): %s",
                ", ".join([f.as_posix() for f in non_exist_files]),
            )
            richprint(
                "[bold bright_red]Files needed for compilation missing (deleted?):[/]"
            )
            for f in non_exist_files:
                richprint(indicator, end=" ")
                print(f"{f!s}")

        if excluded:
            log.debug(
                "Files needed for compilation excluded from tarball: %s",
                ", ".join([f.as_posix() for f in excluded]),
            )
            richprint(
                "[bold bright_red]Files needed for compilation not included:[/]"
            )
            for f in excluded:
                richprint(indicator, end=" ")
                print(f"{f!s}")

        print()  # blank line to section off errors
        return True
    else:
        return False


def _check_warn_extra(
    _ref: Tarballer,
    _tgt: Tarballer,
    _indicator: str = "*",
) -> bool:
    """
    Compares objects included in _ref and _tgt and return True if the
    latter contains files missing from the former. Logs accordingly.
    """
    if _extra := _tgt.objects().difference(_ref.objects()):
        log.debug(
            "Files not essential to compilation added to tarball: %s",
            ", ".join([f.as_posix() for f in _extra]),
        )
        richprint("[yellow]Files not essential to compilation added:[/]")
        for f in _extra:
            richprint(_indicator, end=" ")
            print(f"{f!s}")
        else:
            print()

        return True
    else:
        return False


def _set_main_file(name: str) -> Union[Path, None]:
    """
    Return fully resolved file name after adding an appropriate suffix, if
    necessary. Raises `FileNotFoundError` if neither the input str, nor a
    version of it with '.tex' or '.fls' extension, point to any existing file.
    """
    main_file = Path(name).resolve()
    if main_file.suffix not in [".fls", ".tex"]:
        # ...otherwise try adding the .fls/.tex suffix to main_file
        for f in [str(main_file) + suff for suff in [".fls", ".tex"]]:
            if Path(f).is_file():
                main_file = Path(f)
                break
        else:
            raise FileNotFoundError(
                f"File {main_file.name}[.tex|.fls] not found."
            )
    if main_file.is_file():
        return main_file
    else:
        raise FileNotFoundError(f"File {main_file.name} not found.")


################
## MAIN CLASS ##
################


class TarTeX:
    """
    Class to help build  a tarball including all source files needed to
    re-compile your LaTeX project."
    """

    pkglist_name = "TeXPackages.json"

    def __init__(self, args):
        self.args = parse_args(args)
        log_level = (
            (log.INFO // self.args.verbose)
            if self.args.verbose > 0
            else log.WARN
        )
        log.basicConfig(
            format="%(message)s",
            level=log_level,
            handlers=[
                richlog.RichHandler(
                    show_time=False,
                    show_path=False,
                    highlighter=highlighter.NullHighlighter(),
                )
            ],
        )
        self.cwd = Path.cwd()
        try:
            self.main_file = _set_main_file(self.args.fname)
        except FileNotFoundError as err:
            log.critical(f"Error: {err}")
            sys.exit(1)

        self.filehash_cache = (
            _tartex_xdg_utils.app_cache_dir(self.main_file)
            / f"{self.main_file.stem}.cache"
        )

        if self.args.git_rev:
            try:
                self.GR = GitRev(
                    self.main_file.parent, self.args.git_rev or "HEAD"
                )
                self.tar_file_git_tag = f"-{self.GR.id()}.tar"
            except Exception:
                sys.exit(1)
        else:
            self.tar_file_git_tag = ""

        self.pdf_stream = None
        # ..but use use specified output's TAR_EXT extension if any...
        self.tar_ext = ""
        if self.args.output:
            self.args.output, self.tar_ext = _tartex_tar_utils.proc_output_path(
                self.cwd,
                self.main_file,
                self.args.output,
                self.tar_file_git_tag,
            )

        self.tar_file_w_ext = self._tar_filename()
        self.tar_ext = self.tar_file_w_ext.suffix.lstrip(".")
        log.debug("Output tarball name: '%s'", self.tar_file_w_ext)

        self.req_supfiles = {}
        self.add_files = self.args.add.split(",") if self.args.add else []
        excludes = self.args.excl.split(",") if self.args.excl else []
        excl_lists = (self.main_file.parent.glob(f"**/{L}") for L in excludes)

        self.excl_files = set(
            [
                f.relative_to(self.main_file.parent)
                for L in excl_lists
                for f in L
            ]
        )
        # Do not exclude main tex file even if matched by `--excl` glob
        self.excl_files.discard(
            self.main_file.with_suffix(".tex").relative_to(
                self.main_file.parent
            )
        )

        # If .bbl/.ind is missing in source dir, then it is not in
        # self.excl_files (result of globbing files in srcdir) even if the user
        # passes a matching wildcard to exclude it with "-x".
        # Extend self.excl_files specifically in these cases
        for glb in excludes:
            self.excl_files.update(
                [
                    f
                    for f in [
                        self.main_file.with_suffix(f".{g}")
                        for g in _tartex_tex_utils.SUPP_REQ
                    ]
                    if fnmatch.fnmatch(f.name, glb)
                ]
            )

        if self.excl_files:
            log.debug(" ".join(excl_lists))
            log.info(
                "List of excluded files: %s",
                ", ".join([Path(x).as_posix() for x in self.excl_files]),
            )

        self.force_tex = False if self.args.git_rev else self.args.latexmk_tex
        if not self.force_tex:
            # If force_tex is not set by user options,
            # set to ps if source dir contains ps/eps files
            # or to pdf otherwise
            src_ps = [
                str(p)
                for ext in ["eps", "ps"]
                for p in self.main_file.parent.glob(f"**/*.{ext}")
            ]
            self.force_tex = "ps" if src_ps else "pdf"
            log.info(
                "Latexmk will use %slatex for processing, if needed",
                self.force_tex,
            )

        self.pkglist = None

    def input_files(self, tarf: Tarballer, silent: bool = False) -> None:
        """
        Returns non-system input files needed to compile the main tex file.

        Input files are directly taken from `git ls-tree -r` if `self.git_rev`
        is true. In this case, user specified additional files and exclusions
        are ignored.

        Otherwise, uses `.fls` file, including trying to compile the
        main tex file using `latexmk` if it cannot find the fls file in the
        source dir.
        """

        pkgs: dict[str, set[str]] = {}
        if self.args.git_rev:
            log.debug(
                "Using `git ls-tree` to determine files to include in tarball"
            )
            deps, pkgs = self.input_files_from_git(tarf, silent=silent)

        # If .fls is the user supplied "main file", this assumes that all INPUT
        # files recorded in it are also included in source dir and skips any
        # missing ones (unless force recompiling)
        elif self.main_file.suffix == ".fls" and not self.args.force_recompile:
            deps, pkgs = self.input_files_from_srcfls(tarf, silent=silent)

        # When "main file" is ".tex" and recompile has been forced by user
        # option `--force-recompile`
        elif self.args.force_recompile:
            deps, pkgs = self.input_files_from_recompile(tarf, silent=silent)

        # Most typical case: main file is ".tex", no force recompile, try cache
        # first, recompile if input files are missing or their hash don't match
        else:
            deps, pkgs = self.input_files_from_cache(tarf, silent=silent)

        if self.args.bib:
            tarf.app_files(*self._add_bib(pkgs))

        if self.add_files:
            tarf.app_files(*self._add_user_files())

        if self.excl_files:  # Remove files in excl list
            tarf.drop_files(*self.excl_files)

        if self.args.packages:
            log.info(
                "System TeX/LaTeX packages used: %s",
                ", ".join(sorted(pkgs["System"])),
            )

            self.pkglist = json.dumps(
                pkgs, cls=_tartex_tex_utils.SetEncoder
            ).encode("utf8")
            tarf.app_stream(
                self.pkglist_name,
                self.pkglist,
                comm=f"Adding list of used LaTeX packages: {self.pkglist_name}",
            )

    def input_files_from_git(
        self, tarf: Tarballer, silent: bool = False
    ) -> tuple[set[Path], dict[str, set[str]]]:
        """
        Get input files from `git ls-tree <REVISION>`
        :returns: TODO

        """
        if not silent:
            log.info("Getting files from git-tree")

        _deps = self.GR.ls_tree_files()
        _pkgs: dict[str, set[str]] = {}
        # Process the --excl files
        tarf.app_files(*_deps)
        tarf.set_mtime(self.GR.mtime())
        if self.args.packages:
            try:
                with open(
                    self.main_file.with_suffix(".fls"), encoding="utf-8"
                ) as f:
                    _, _pkgs = _latex.fls_input_files(
                        f,
                        _tartex_tex_utils.AUXFILES,
                        sty_files=self.args.packages,
                    )
            except FileNotFoundError:
                if not silent:
                    log.warning(
                        "Missing .fls file in source dir; %s will not be saved",
                        self.pkglist_name,
                    )
                self.args.packages = False
            except Exception as err:
                raise err
        return _deps, _pkgs

    def input_files_from_recompile(
        self,
        tarf: Tarballer,
        minimal: bool = False,
        cache_update: bool = False,
        silent: bool = False,
    ) -> tuple[set[Path], dict[str, set[str]]]:
        """
        Determines set of input files and list of LaTeX packages used by
        running explicit re-compilation of project in an out-of-tree temporary
        directory.

        If minimal is `True`, omits adding any files other than those
        absolutely needed for compilation, such as user specified bib files,
        etc., or, likewise, excluding files matching user specified exclusion
        patterns if any.

        Returns tuple: (Set of Path objects representing input files, dict of
        LaTeX packages used for compilation)
        """
        _deps: set[Path] = set()
        _pkgs: dict[str, set[str]] = {}
        with TemporaryDirectory() as compile_dir:
            if not silent:
                log.info(
                    f"LaTeX recompile {'forced' if self.args.force_recompile else 'required'}"
                )
                log.info("LaTeX compile directory: %s", compile_dir)
            try:
                fls_path = _latex.run_latexmk(
                    self.main_file.with_suffix(".tex"),
                    self.force_tex,
                    compile_dir,
                    silent=silent,
                )
            except Exception as e:
                raise e

            tarf.set_mtime(os.path.getmtime(fls_path))
            with open(fls_path, encoding="utf-8") as f:
                _deps, _pkgs = _latex.fls_input_files(
                    f,
                    _tartex_tex_utils.AUXFILES,
                    sty_files=self.args.packages,
                )
                tarf.app_files(*_deps)

                # Supplementary files, pkg lists are all transient — they only
                # live in the temporary `compile_dir`; so we need to save them
                # as streams inside this file-open context.
                self._add_supplement_streams(Path(compile_dir), _deps, tarf)
                if not minimal:
                    if self.args.with_pdf:
                        self._add_pdf_stream(fls_path.with_suffix(".pdf"), tarf)

                if cache_update:
                    _tartex_hash_utils.save_input_files_hash(
                        Path(self.filehash_cache),
                        _deps,
                        tarf.streams(),
                        _pkgs
                    )
                return _deps, _pkgs

    def input_files_from_srcfls(
        self, _t: Tarballer, silent: bool = False
    ) -> tuple[set[Path], dict[str, set[str]]]:
        """Get input files from '.fls' in source dir"""

        if not silent:
            log.info(
                "Reading input files from project fls file: '%s'",
                self.main_file.name
            )

        _deps: set[Path] = set()
        _pkgs: dict[str, set[str]] = {}

        fls_f = self.main_file.with_suffix(".fls")
        try:
            with open(fls_f, encoding="utf8") as f:
                _deps, _pkgs = _latex.fls_input_files(
                    f,
                    _tartex_tex_utils.AUXFILES,
                    sty_files=self.args.packages,
                )
                _t.set_mtime(os.path.getmtime(fls_f))

        except FileNotFoundError:
            if self.args.packages:
                if not silent:
                    log.warning(
                        "Cannot generate list of packages due to missing %s file",
                        fls_f,
                    )
                self.args.packages = False

        except Exception as err:
            raise err

        _t.app_files(*_deps)

        if self.args.with_pdf:
            self._add_pdf_stream(self.main_file.with_suffix(".pdf"), _t)
        return _deps, _pkgs

    def input_files_from_cache(
        self, _t: Tarballer, silent: bool = False
    ) -> tuple[set[Path], dict[str, set[str]]]:
        """
        Determine input files based on cache data:
        * If no cache data exists, recompile and store cache data.
        * If cache data exists and input files listed in it have their sha sums
          matching those of files in project — indicating no change in input
          files content, directly use these input files from project dir.
        * If cache data exists, but input files are missing or sha sums do not
          match up, recompile and update cache data.

        Returns: tuple of input dependency files paths, dict corresponding to
        package list
        """
        cache_path = Path(self.filehash_cache)
        try:
            c_path_msg = f"~/{cache_path.relative_to(Path.home())}"
        except ValueError:
            c_path_msg = str(cache_path)
        if not silent:
            log.info("Reading cache file: %s", c_path_msg)

        if cache_path.is_file():
            _deps: set[Path] = set()
            _pkgs: dict[str, set[str]] = {}
            if _tartex_hash_utils.check_file_hash(cache_path):
                if not silent:
                    log.info("No changes to input files, using cached data")
                with open(cache_path, "r") as cache:
                    _j = json.load(cache)
                    _deps.update([Path(f) for f in _j["input_files"].keys()])
                    _pkgs = _j["packages"]
                    missing: set[Path] = set()  # Missing files to discard

                    # Apart from obviously missing files from the set of _deps,
                    # anything in "streams" is automatically assumed missing
                    # since there is no way to have it restored and still
                    # verify its integrity. Resolved upon next cache update.
                    for _d in _deps.union([Path(_f) for _f in _j["streams"]]):
                        if not _d.is_file():
                            log.warning(
                                "Missing input file %s, try recompile ('-F')",
                                _d.name,
                            )
                            missing.add(_d)
                    _deps = _deps.difference(missing)
                _t.app_files(*_deps)
                if self.args.with_pdf:
                    self._add_pdf_stream(
                        self.main_file.with_suffix(".pdf"), _t, silent
                    )
                return _deps, _pkgs
            else:
                if not silent:
                    log.info("Dependencies missing or modified...")
        else:
            if not silent:
                log.info("Cache file missing...")

        return self.input_files_from_recompile(_t, cache_update=True)

    def _add_bib(self, _pkgs: dict[str, set[str]]):
        """
        Add bib and bst files to tarball; add bst filename to package list

        :_deps: set of files to be included in tarball
        :_pkgs: list of TeX packages that may be added to tarball as json file

        """
        bibs = set(
            [
                f
                for f in _tartex_tex_utils.bib_file(
                    self.main_file.with_suffix(".tex")
                )
                if f
            ]
        )

        for f in bibs:
            try:
                if f and re.match(r".bst", f.suffix):
                    _pkgs["Local"].add(f.as_posix())
            except Exception:
                pass

        return bibs

    def _add_user_files(self):
        """Add user specified extra files to tarball"""
        _files = set()
        for f in _tartex_tex_utils.add_files(
            self.add_files, self.main_file.parent
        ):
            f_relpath = f.relative_to(self.main_file.parent)
            _files.add(f_relpath)
            log.info("Add user specified file: %s", f_relpath)
        return _files

    def _add_pdf_stream(self, _file: Path, _t: Tarballer, silent: bool = False):
        """
        Add pdf as stream to Tarballer object

        :_file: path to pdf file
        :_t: Tarballer object to append pdf stream to
        :silent: suppress logs when `True` (bool)
        :returns: None

        """
        try:
            with open(_file, "rb") as f:
                self.pdf_stream = f.read()
                _t.app_stream(_file.name, self.pdf_stream)
        except FileNotFoundError:
            if not silent:
                log.warning(
                    "Skipping pdf not found: %s",
                    f"{_file.relative_to(self.main_file.parent)}",
                )
            self.args.with_pdf = False

    def _add_supplement_streams(self, _p: Path, _dep: set[Path], _t: Tarballer):
        """
        Add supplementary files as streams

        :_p: Path to stream
        :_t: Tarballer object to add `_p` to

        """
        for ext in _tartex_tex_utils.SUPP_REQ:
            supp_filename = self.main_file.with_suffix(f".{ext}")
            if app := self._missing_supp(
                self.main_file.with_suffix(f".{ext}"), _p, _dep
            ):
                self.req_supfiles[self.main_file.with_suffix(f".{ext}")] = app
                _t.app_stream(supp_filename.name, app)

    def tar_files(self):
        """
        Generates a tarball consisting of non-system input files needed to
        recompile your latex project. Main entry point to tartex cmdline app.
        """
        _git_cntxt = (
            git_checkout(self.GR.git_bin, self.GR.repo, self.GR.rev)
            if self.args.git_rev
            else nullcontext()
        )
        _pushd_src_dir = chdir(self.main_file.parent)
        if self.args.check or self.args.only_check:
            try:
                with _git_cntxt:
                    with _pushd_src_dir:
                        self.check_files(not self.args.only_check)
            except Exception:
                if not self.args.only_check:
                    log.critical("Check failed, no tarball will be generated")
                sys.exit(1)

        if self.args.only_check:
            return

        if self.tar_file_w_ext.exists() and not self.args.list:
            self.tar_file_w_ext, self.tar_ext = (
                _tartex_tar_utils.tar_name_conflict(
                    self.cwd,
                    self.main_file,
                    self.tar_file_w_ext,
                    self.args.overwrite,
                    self.tar_file_git_tag,
                )
            )
        tarball = Tarballer(self.cwd, self.main_file, self.tar_file_w_ext)
        try:
            with _git_cntxt:
                with _pushd_src_dir:
                    log.info(
                        "Switched to TeX file source dir: %s",
                        self.main_file.parent,
                    )
                    _ = self.input_files(tarball)
                    if self.args.list:
                        log.debug(
                            "Working in 'list' mode; no tarball will be produced"
                        )
                        tarball.print_list(self.args.summary)
                    else:
                        tarball.do_tar()
                        if self.args.summary:
                            _tartex_msg_utils.summary_msg(
                                tarball.num_objects,
                                self.tar_file_w_ext,
                                self.cwd,
                            )
                    log.info("Switching back to working dir: %s", self.cwd)
        except Exception:
            sys.exit(1)

    def check_files(self, silent: bool = False):
        """
        Checks if simulated target tarball will contain all files needed to
        recompile project (and any more).
        """

        log.debug("Working in `check` mode; no tarball will be produced")
        log.debug("Checking whether target tarball contents will recompile...")
        richprint(
            "[bright_black]Checking whether target tarball contents will recompile...[/]",
        )

        # Rich print indicators for missing or superfluous file messages
        INDI = {
            "req-miss": ":double_exclamation_mark-emoji: ",  # extra space req
            "not-need": ":warning-emoji: ",  # extra space req
            "perfect": ":heavy_check_mark-emoji: ",  # extra space req
            "chk-fail": ":cross_mark-emoji:",
            "chk-pass": ":ok_button-emoji:",
        }

        # The reference tarball `ref_tar` must contain the minimal number of
        # files/streams required to re-compile the project, ignoring any user
        # specified additionals. Thus, we determine the objects that will go
        # into this explicit recompilation (if latexmk fails at this stage,
        # the check is assumed to have failed — we cannot work around this) by
        # using `.input_files_from_recompile()`.
        #
        # When comparing a target tarfile against this ref, any additional
        # files in the former would raise a warning, but any missing file is an
        # error.
        #
        ref_tar = Tarballer(self.cwd, self.main_file, Path("ref.tar"))
        try:
            # passing `minimal=True` ensures tarballer obj includes only the
            # absolutely necessary files/streams as determined from a recompile
            deps, pkgs = self.input_files_from_recompile(
                ref_tar, minimal=True, silent=silent
            )
        except Exception as err:
            log.critical(
                "Latexmk failed to compile; check if all source files exist"
            )
            if not silent:
                richprint(
                    f"{INDI['req-miss']} Files need for compilation missing (deleted?)"
                )
            raise CheckFailError from err

        # dummy target including all user specified additionals/exclusions
        dummy_tar = Tarballer(self.cwd, self.main_file, Path("dummy.tar"))
        miss_msg = (
            "Files needed for compilation missing or excluded from tarball"
        )

        missing_supp = (
            set(
                [
                    self.main_file.with_suffix(f".{e}").relative_to(
                        self.main_file.parent
                    )
                    for e in _tartex_tex_utils.SUPP_REQ
                ]
            )
            if self.args.force_recompile
            else set()
        )
        # make sure supplementary file is actually needed
        missing_supp.intersection_update(ref_tar.objects())

        if not self.args.git_rev:
            self.input_files(dummy_tar, silent)

            # will return True if check fails
            chk_err = _check_err_missing(
                ref_tar,
                dummy_tar,
                indicator=INDI["req-miss"],
                supp_files=missing_supp,
            )
            if not silent:
                _ = _check_warn_extra(ref_tar, dummy_tar, INDI["not-need"])

        else:
            git_ref = self.args.git_rev or "HEAD"
            # Dummy target; will be populated by contents from `git ls-tree`
            # (plus user additionals)
            self.input_files(dummy_tar, silent)

            chk_err = _check_err_missing(
                ref_tar,
                dummy_tar,
                indicator=INDI["req-miss"],
                supp_files=missing_supp,
            )
            if not silent:
                _ = _check_warn_extra(ref_tar, dummy_tar, INDI["not-need"])

        if silent:
            if chk_err:
                log.debug("Check failed")
                richprint("[red]Check failed[/]")
            else:
                log.debug("Check succeeded")
                richprint("[green4]Check success[/]")
            if missing_supp:
                # log INFO if force recompiling, WARNING otherwise
                log.log(
                    log.INFO if self.args.force_recompile else log.WARNING,
                    "Missing supplementary file(s): %s",
                    ", ".join([str(_s) for _s in missing_supp]),
                )
        else:
            richprint("[green4]Files needed for compilation to be included:[/]")
            for f in ref_tar.objects().intersection(dummy_tar.objects()):
                richprint(INDI["perfect"], end=" ")
                print(f"{f!s}")
            else:
                print()

            if chk_err:
                richprint(f"{INDI['chk-fail']} [red]{miss_msg}[/]")
            else:
                log.info("All files needed for compilation are present")
                richprint(
                    f"{INDI['chk-pass']} [bold green4]All files needed for"
                    " compilation included in tarball"
                    f"{f' at git revision {git_ref}' if self.args.git_rev else ''}[/]"
                )

        if chk_err:
            raise CheckFailError(miss_msg)

    ############################
    ## CLASS HELPER FUNCTIONS ##
    ############################

    def _missing_supp(self, fpath, tmpdir, deps):
        """Handle missing supplementary file from orig dir, if req"""
        if (
            fpath not in deps  # bbl file not in source dir
            and (Path(tmpdir) / fpath.name).exists()  # Implies it's req
            and fpath not in self.excl_files  # Not explicitly excluded
        ):
            log.debug("Required file '%s' not in source dir", fpath.name)
            log.info(
                "Add contents as BytesIO: %s",
                Path(tmpdir) / fpath.name,
            )
            return (Path(tmpdir) / fpath.name).read_bytes()

        return None

    def _tar_filename(self):
        """
        Set member variables for tar output
        :returns: full tarball file name

        """
        _tar_ext = (
            self.tar_ext if self.tar_ext else _tartex_tar_utils.TAR_DEFAULT_COMP
        )
        # ...but overwrite TAR_EXT if a specific tar compression option passed
        if self.args.bzip2:
            _tar_ext = "bz2"
        if self.args.gzip:
            _tar_ext = "gz"
        if self.args.xz:
            _tar_ext = "xz"

        tar_file = (
            Path(f"{self.args.output}.tar")
            if self.args.output
            else Path(
                f"{self.main_file.stem}{self.tar_file_git_tag}"
            ).with_suffix(".tar")
        )

        # Note: if tar_file is already an abs path, 'foo/tar_file' simply returns 'tar_file'
        return self.cwd / tar_file.with_suffix(f".tar.{_tar_ext}")


###################################
## COMMAND LINE UTILITY FUNCTION ##
###################################
def make_tar():
    """Build tarball with command line arguments processed by argparse"""
    t = TarTeX(sys.argv[1:])
    t.tar_files()


if __name__ == "__main__":
    make_tar()
